"""
Module for snow masking

Credits:
Copyright (c) 2017-2019 Matej Aleksandrov, Matej Batič, Andrej Burja, Eva Erzin (Sinergise)
Copyright (c) 2017-2019 Grega Milčinski, Matic Lubej, Devis Peresutti, Jernej Puc (Sinergise)
Copyright (c) 2017-2019 Jovan Višnjić, Anže Zupanc, Lojze Žust (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""

import logging
import itertools

import numpy as np
from skimage.morphology import disk, binary_dilation

from eolearn.core import EOTask, FeatureType
from .utilities import resize_images


LOGGER = logging.getLogger(__name__)


class BaseSnowMask(EOTask):
    """ Base class for snow detection and masking
    """
    def __init__(self, data_feature, band_indices, dilation_size=0, undefined_value=0, mask_name='SNOW_MASK'):
        """
        :param data_feature: EOPatch feature represented by a tuple in the form of `(FeatureType, 'feature_name')`
        :type data_feature: tuple(FeatureType, str)
        :param band_indices: A list containing the indices at which the required bands can be found in the data_feature.
        :type band_indices: list(int)
        :param dilation_size: Size of the disk in pixels for performing dilation. Value 0 means do not perform
            this post-processing step.
        :type dilation_size: int
        """
        self.bands_feature = next(self._parse_features(data_feature)())
        self.band_indices = band_indices
        self.dilation_size = dilation_size
        self.undefined_value = undefined_value
        self.mask_feature = (FeatureType.MASK, mask_name)

    def _apply_dilation(self, snow_masks):
        """ Apply binary dilation for each mask in the series """
        if self.dilation_size:
            snow_masks = np.array([binary_dilation(mask, disk(self.dilation_size)) for mask in snow_masks])
        return snow_masks

    def execute(self, eopatch):
        raise NotImplementedError


class SnowMask(BaseSnowMask):
    """ The task calculates the snow mask using the given thresholds.

    The default values were optimised based on the Sentinel-2 L1C processing level. Values might not be optimal for L2A
    processing level
    """

    NDVI_THRESHOLD = 0.1

    def __init__(self, data_feature, band_indices, ndsi_threshold=0.4, brightness_threshold=0.3, **kwargs):
        """
        :param data_feature: EOPatch feature represented by a tuple in the form of `(FeatureType, 'feature_name')`
            containing the bands 2, 3, 7, 11, i.e. (FeatureType.DATA, 'BANDS')
        :type data_feature: tuple(FeatureType, str)
        :param band_indices: A list containing the indices at which the required bands can be found in the data_feature.
            The required bands are B03, B04, B08 and B11 and the indices should be provided in this order. If the
            'BANDS' array contains all 13 L1C bands, then `band_indices=[2, 3, 7, 11]`. If the 'BANDS' are the 12 bands
            with L2A values, then `band_indices=[2, 3, 7, 10]`
        :type band_indices: list(int)
        :param ndsi_threshold: Minimum value of the NDSI required to classify the pixel as snow
        :type ndsi_threshold: float
        :param brightness_threshold: Minimum value of the red band for a pixel to be classified as bright
        :type brightness_threshold: float
        """
        super().__init__(data_feature, band_indices, **kwargs)
        self.ndsi_threshold = ndsi_threshold
        self.brightness_threshold = brightness_threshold

    def execute(self, eopatch):
        bands = eopatch[self.bands_feature][..., self.band_indices]
        with np.errstate(divide='ignore'):
            # (B03 - B11) / (B03 + B11)
            ndsi = (bands[..., 0] - bands[..., 3]) / (bands[..., 0] + bands[..., 3])
            # (B08 - B04) / (B08 + B04)
            ndvi = (bands[..., 2] - bands[..., 1]) / (bands[..., 2] + bands[..., 1])

        ndsi_invalid, ndvi_invalid = ~np.isfinite(ndsi), ~np.isfinite(ndvi)
        ndsi[ndsi_invalid] = self.undefined_value
        ndvi[ndvi_invalid] = self.undefined_value

        snow_mask = np.where(np.logical_and(np.logical_or(ndsi >= self.ndsi_threshold,
                                                          np.abs(ndvi - self.NDVI_THRESHOLD) < self.NDVI_THRESHOLD / 2),
                                            bands[..., 0] >= self.brightness_threshold), 1, 0)

        snow_mask = self._apply_dilation(snow_mask)

        snow_mask[np.logical_or(ndsi_invalid, ndvi_invalid)] = self.undefined_value

        eopatch[self.mask_feature] = snow_mask[..., np.newaxis].astype(bool)
        return eopatch


class TheiaSnowMask(BaseSnowMask):
    """ Task to add a snow mask to an EOPatch. The input data is either Sentinel-2 L1C or L2A level

    Original implementation and documentation available at https://gitlab.orfeo-toolbox.org/remote_modules/let-it-snow

    ATBD https://gitlab.orfeo-toolbox.org/remote_modules/let-it-snow/blob/master/doc/atbd/ATBD_CES-Neige.pdf

    This task computes a snow mask for the input EOPatch. The `data_feature` to be used as input to the
    classifier is a mandatory argument. If all of the needed features exist already, the classifier is run.
    `linear` interpolation is used for resampling of the `data_feature` and cloud probability map, while `nearest`
    interpolation is used to upsample the binary cloud mask.
    """
    B10_THR = 0.015
    DEM_FACTOR = 0.00001

    def __init__(self, data_feature, band_indices, cloud_mask_feature, dem_feature, dem_params=(100, 0.1),
                 red_params=(12, 0.3, 0.1, 0.2, 0.040), ndsi_params=(0.4, 0.15, 0.001), b10_index=None, **kwargs):
        """
        :param data_feature: EOPatch feature represented by a tuple in the form of `(FeatureType, 'feature_name')`
            containing the bands B3, B4, and B11

            Example: `(FeatureType.DATA, 'ALL-BANDS')`
        :type data_feature: tuple(FeatureType, str)
        :param band_indices: A list containing the indices at which the required bands can be found in the bands
            feature. If all L1C band values are provided, `band_indices=[2, 3, 11]`. If all L2A band values are
            provided, then `band_indices=[2, 3, 10]`
        :type band_indices: list(int)
        :param cloud_mask_feature: EOPatch CLM feature represented by a tuple in the form of
            `(FeatureType, 'feature_name')` containing the cloud mask
        :type cloud_mask_feature: tuple(FeatureType, str)
        :param dem_feature: EOPatch DEM feature represented by a tuple in the form of `(FeatureType, 'feature_name')`
            containing the digital elevation model
        :type dem_feature: tuple(FeatureType, str)
        :param b10_index: Array index where the B10 band is stored in the bands feature. This is used to refine the
            initial cloud mask
        :type b10_index: int
        :param dem_params: Tuple with parameters pertaining DEM processing. The first value specifies the bin size
            used to group DEM values, while the second value specifies the minimum snow fraction in an elevation band
            to define z_s. With reference to the ATBD, the tuple is (d_z, f_t)
        :type dem_params: (float, float)
        :param red_params: Tuple specifying parameters to process the B04 red band. The first parameter defines the
            scaling factor for down-sampling the red band, the second parameter is the maximum value of the
            down-sampled red band for a dark cloud pixel, the third parameter is the minimum value
            to return a non-snow pixel to the cloud mask, the fourth is the minimum reflectance value to pass the 1st
            snow test, and the fifth is the minimum reflectance value to pass the 2nd snow test. With reference to the
            ATBD, the tuple is (r_f, r_d, r_b, r_1, r_2)
        :type red_params: (float, float, float, float, float)
        :param ndsi_params: Tuple specifying parameters for the NDSI. First parameter is the minimum value to pass the
            1st snow test, the second parameter is the minimum value to pass the 2nd snow test, and the third parameter
            is the minimum snow fraction in the image to activate the pass 2 snow test. With reference to the
            ATBD, the tuple is (n_1, n_2, f_s)
        :type ndsi_params: (float, float, float)
        """
        super().__init__(data_feature, band_indices, **kwargs)
        self.dem_feature = next(self._parse_features(dem_feature)())
        self.clm_feature = next(self._parse_features(cloud_mask_feature)())
        self.dem_params = dem_params
        self.red_params = red_params
        self.ndsi_params = ndsi_params
        self.b10_index = b10_index
        self._validate_params()

    def _validate_params(self):
        """ Check length of parameters defining threshold values """
        for params, n_params in [(self.dem_params, 2), (self.red_params, 5), (self.ndsi_params, 3)]:
            if not isinstance(params, (tuple, list)) or len(params) != n_params:
                raise ValueError('Incorrect format or number of parameters for {0}. '
                                 'Has to be a tuple of length {1}'.format(params, n_params))

    def _resample_red(self, input_array):
        """ Method to resample the values of the red band

        The input array is first down-scaled using bicubic interpolation and up-scaled back using nearest neighbour
        interpolation

        :param input_array: input values
        :return: resampled values
        """
        height, width = input_array.shape[1:]
        size = (height // self.red_params[0], width // self.red_params[0])
        return resize_images(resize_images(input_array[..., np.newaxis], new_size=size),
                             new_size=(height, width)).squeeze()

    def _adjust_cloud_mask(self, bands, cloud_mask, dem, b10):
        """ Adjust existing cloud mask using cirrus band if L1C data and resampled red band

        Add to the existing cloud mask pixels found thresholding down-sampled red band and cirrus band/DEM
        """
        clm_b10 = np.where(b10 > self.B10_THR + self.DEM_FACTOR * dem, 1, 0) \
            if b10 is not None else np.ones(shape=cloud_mask.shape, dtype=np.uint8)
        return np.logical_or(
            np.where(np.logical_and(cloud_mask == 1, self._resample_red(bands[..., 1]) > self.red_params[1]), 1, 0),
            clm_b10).astype(np.uint8)

    def _apply_first_pass(self, bands, ndsi, clm, dem, clm_temp):
        """ Apply first pass of snow detection """
        snow_mask_pass1 = np.where(np.logical_and(np.logical_not(clm_temp),
                                                  np.logical_and(ndsi > self.ndsi_params[0],
                                                                 bands[..., 1] > self.red_params[3])), 1, 0)

        clm_pass1 = np.where(
            np.logical_or(clm_temp, (bands[..., 1] > self.red_params[2]) &
                          np.logical_not(snow_mask_pass1) & clm), 1, 0)

        dem_edges = np.linspace(np.min(dem), np.max(dem),
                                int(np.ceil((np.max(dem) - np.min(dem)) / self.dem_params[0])))
        nbins = len(dem_edges) - 1
        dem_hist_clear_pixels, snow_frac = None, None
        if nbins > 0:
            snow_frac = np.zeros(shape=(bands.shape[0], nbins))
            dem_hist_clear_pixels = np.array([np.histogram(dem[np.logical_not(mask)], bins=dem_edges)[0]
                                              for mask in clm_pass1])

            for date, nbin in itertools.product(range(bands.shape[0]), range(nbins)):
                if dem_hist_clear_pixels[date, nbin] > 0:
                    dem_mask = np.logical_and(dem_edges[nbin] <= dem, dem < dem_edges[nbin + 1])
                    in_dem_range_clear = np.where(
                        np.logical_and(dem_mask, np.logical_not(clm_pass1[date])))
                    snow_frac[date, nbin] = \
                        np.sum(snow_mask_pass1[date][in_dem_range_clear]) / dem_hist_clear_pixels[date, nbin]
        return snow_mask_pass1, snow_frac, dem_edges

    def _apply_second_pass(self, bands, ndsi, dem, clm_temp, snow_mask_pass1, snow_frac, dem_edges):
        """ Second pass of snow detection """
        _, height, width, _ = bands.shape
        total_snow_frac = np.sum(snow_mask_pass1, axis=(1, 2)) / (height*width)
        snow_mask_pass2 = np.zeros(snow_mask_pass1.shape)
        for date in range(bands.shape[0]):
            if (total_snow_frac[date] > self.ndsi_params[2]) and \
                    (snow_frac is not None and np.any(snow_frac[date] > self.dem_params[1])):
                z_s = dem_edges[max(np.argmax(snow_frac[date] > self.dem_params[1]) - 2, 0)]
                snow_mask_pass2[date, :, :] = np.where(
                    np.logical_and(dem > z_s, np.logical_and(
                        np.logical_not(clm_temp[date]),
                        np.logical_and(ndsi[date] > self.ndsi_params[1],
                                       bands[date, ..., 1] > self.red_params[-1]))), 1, 0)
        return snow_mask_pass2

    def execute(self, eopatch):
        """ Run multi-pass snow detection """
        bands = eopatch[self.bands_feature][..., self.band_indices]
        b10 = eopatch[self.bands_feature][..., self.b10_index] if self.b10_index is not None else None
        dem = eopatch[self.dem_feature][..., 0]
        clm = eopatch[self.clm_feature][..., 0]

        with np.errstate(divide='ignore'):
            # (B03 - B11) / (B03 + B11)
            ndsi = (bands[..., 0] - bands[..., 2]) / (bands[..., 0] + bands[..., 2])

        ndsi_invalid = ~np.isfinite(ndsi)
        ndsi[ndsi_invalid] = self.undefined_value

        clm_temp = self._adjust_cloud_mask(bands, clm, dem, b10)

        snow_mask_pass1, snow_frac, dem_edges = self._apply_first_pass(bands, ndsi, clm, dem, clm_temp)

        snow_mask_pass2 = self._apply_second_pass(bands, ndsi, dem, clm_temp, snow_mask_pass1, snow_frac, dem_edges)

        snow_mask = self._apply_dilation(np.logical_or(snow_mask_pass1, snow_mask_pass2))

        eopatch[self.mask_feature] = snow_mask[..., np.newaxis].astype(bool)

        return eopatch



"""
Module for cloud masking

Credits:
Copyright (c) 2017-2019 Matej Aleksandrov, Matej Batič, Andrej Burja, Eva Erzin (Sinergise)
Copyright (c) 2017-2019 Grega Milčinski, Matic Lubej, Devis Peresutti, Jernej Puc, Tomislav Slijepčević (Sinergise)
Copyright (c) 2017-2019 Blaž Sovdat, Nejc Vesel, Jovan Višnjić, Anže Zupanc, Lojze Žust (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""

import logging
import os
import warnings

import cv2
import numpy as np
from lightgbm import Booster
from eolearn.core import EOTask, FeatureType, execute_with_mp_lock
from sentinelhub import bbox_to_resolution
from skimage.morphology import disk

from .utilities import map_over_axis, resize_images

LOGGER = logging.getLogger(__name__)


class CloudMaskTask(EOTask):
    """ Cloud masking with an improved s2cloudless model and the SSIM-based multi-temporal classifier.

    Its intended output is a cloud mask that is based on the outputs of both
    individual classifiers (a dilated intersection of individual binary masks).
    Additional cloud masks and probabilities can be added for either classifier or both.

    Prior to feature extraction and classification, it is recommended that the input be
    downscaled by specifying the source and processing resolutions. This should be done
    for the following reasons:

    - faster execution
    - lower memory consumption
    - noise mitigation

    Resizing is performed with linear interpolation. After classification, the cloud
    probabilities are themselves upscaled to the original dimensions, before proceeding
    with masking operations.

    Example usage:

    .. code-block:: python

        # Only output the combined mask
        task1 = AddMultiCloudMaskTask(processing_resolution='120m',
                                      mask_feature='CLM_INTERSSIM',
                                      average_over=16,
                                      dilation_size=8)

        # Only output monotemporal masks. Only monotemporal processing is done.
        task2 = AddMultiCloudMaskTask(processing_resolution='120m',
                                      mono_features=(None, 'CLM_S2C'),
                                      mask_feature=None,
                                      average_over=16,
                                      dilation_size=8)
    """

    MODELS_FOLDER = os.path.join(os.path.dirname(__file__), 'models')
    MONO_CLASSIFIER_NAME = 'pixel_s2_cloud_detector_lightGBM_v0.2.txt'
    MULTI_CLASSIFIER_NAME = 'ssim_s2_cloud_detector_lightGBM_v0.2.txt'

    def __init__(self,
                 data_feature='BANDS-S2-L1C',
                 is_data_feature='IS_DATA',
                 all_bands=True,
                 processing_resolution=None,
                 max_proc_frames=11,
                 mono_features=None,
                 multi_features=None,
                 mask_feature='CLM_INTERSSIM',
                 mono_threshold=0.4,
                 multi_threshold=0.5,
                 average_over=4,
                 dilation_size=2,
                 mono_classifier=None,
                 multi_classifier=None):
        """
        :param data_feature: A data feature which stores raw Sentinel-2 reflectance bands.
            Default value: `'BANDS-S2-L1C'`.
        :type data_feature: str
        :param is_data_feature: A mask feature which indicates whether data is valid.
            Default value: `'IS_DATA'`.
        :type is_data_feature: str
        :param all_bands: Flag, which indicates whether images will consist of all 13 Sentinel-2 bands or only
            the required 10. Default value:  `True`.
        :type all_bands: bool
        :param processing_resolution: Resolution to be used during the computation of cloud probabilities and masks,
            expressed in meters. Resolution is given as a pair of x and y resolutions. If a single value is given,
            it is used for both dimensions. Default is `None` (source resolution).
        :type processing_resolution: int or (int, int)
        :param max_proc_frames: Maximum number of frames (including the target, for multi-temporal classification)
            considered in a single batch iteration (To keep memory usage at agreeable levels, the task operates on
            smaller batches of time frames). Default value:  `11`.
        :type max_proc_frames: int
        :param mono_features: Tuple of keys to be used for storing cloud probabilities and masks (in that order!) of
            the mono classifier. The probabilities are added as a data feature, while masks are added as a mask
            feature. By default, none of them are added.
        :type mono_features: (str or None, str or None)
        :param multi_features: Tuple of keys used for storing cloud probabilities and masks of the multi classifier.
            The probabilities are added as a data feature, while masks are added as a mask feature. By default,
            none of them are added.
        :type multi_features: (str or None, str or None)
        :param mask_feature: Name of the output intersection feature. The masks are added to the `eopatch.mask`
            attribute dictionary. Default value: `'CLM_INTERSSIM'`. If None, the intersection feature is not computed.
        :type mask_feature: str or None
        :param mono_threshold: Cloud probability threshold for the mono classifier. Default value: `0.4`.
        :type mono_threshold: float
        :param multi_threshold: Cloud probability threshold for the multi classifier. Default value: `0.5`.
        :type multi_threshold: float
        :param average_over: Size of the pixel neighbourhood used in the averaging post-processing step.
            A value of `0` or `None` skips this post-processing step. Default value mimics the default for
            s2cloudless: `4`.
        :type average_over: int or None
        :param dilation_size: Size of the dilation post-processing step. A value of `0` or `None` skips this
            post-processing step. Default value mimics the default for s2cloudless: `2`.
        :type dilation_size: int or None
        :param mono_classifier: Classifier used for mono-temporal cloud detection (`s2cloudless` or equivalent).
            Must work on the 10 selected reflectance bands as features `("B01", "B02", "B04", "B05", "B08", "B8A",
            "B09", "B10", "B11", "B12")`. Default value: `None` (s2cloudless is used)
        :type mono_classifier: lightgbm.Booster or sklearn.base.BaseEstimator
        :param multi_classifier: Classifier used for multi-temporal cloud detection.
            Must work on the 90 multi-temporal features:

            - raw reflectance value in the target frame,
            - average value within a spatial window in the target frame,
            - maximum, mean and standard deviation of the structural similarity (SSIM)
            - indices between a spatial window in the target frame and every other,
            - minimum and mean reflectance of all available time frames,
            - maximum and mean difference in reflectances between the target frame and every other.

            Default value: None (SSIM-based model is used)
        :type multi_classifier: lightgbm.Booster or sklearn.base.BaseEstimator
        """
        self.proc_resolution = self._parse_resolution_arg(processing_resolution)

        self._mono_classifier = mono_classifier
        self._multi_classifier = multi_classifier

        self.data_feature = self._parse_features(data_feature, default_feature_type=FeatureType.DATA)
        self.is_data_feature = self._parse_features(is_data_feature, default_feature_type=FeatureType.MASK)
        self.band_indices = (0, 1, 3, 4, 7, 8, 9, 10, 11, 12) if all_bands else tuple(range(10))

        self.max_proc_frames = max_proc_frames

        if mono_features is not None and isinstance(mono_features, tuple):
            self.mono_features = mono_features
        else:
            self.mono_features = (None, None)

        if multi_features is not None and isinstance(multi_features, tuple):
            self.multi_features = multi_features
        else:
            self.multi_features = (None, None)

        if mask_feature:
            mask_feature = next(self._parse_features(mask_feature, default_feature_type=FeatureType.MASK)())
        self.mask_feature = mask_feature

        self.mono_threshold = mono_threshold
        self.multi_threshold = multi_threshold

        if average_over is not None and average_over > 0:
            self.avg_kernel = disk(average_over) / np.sum(disk(average_over))
        else:
            self.avg_kernel = None

        if dilation_size is not None and dilation_size > 0:
            self.dil_kernel = disk(dilation_size).astype(np.uint8)
        else:
            self.dil_kernel = None

        # Because the real sigma is calculated only later in execute method this task is not thread safe!
        self.sigma = 1.

    @staticmethod
    def _parse_resolution_arg(res):
        """ Parses initialization resolution argument
        """
        if res is None:
            return None

        if isinstance(res, (int, float, str)):
            res = res, res

        if isinstance(res, tuple) and len(res) == 2:
            return tuple(float(rs.strip('m')) if isinstance(rs, str) else rs for rs in res)

        raise ValueError("Wrong resolution parameter passed as an argument.")

    @property
    def mono_classifier(self):
        """ An instance of pre-trained mono-temporal cloud classifier. It is loaded only the first time it is required.
        """
        if self._mono_classifier is None:
            path = os.path.join(self.MODELS_FOLDER, self.MONO_CLASSIFIER_NAME)
            self._mono_classifier = Booster(model_file=path)

        return self._mono_classifier

    @property
    def multi_classifier(self):
        """ An instance of pre-trained multi-temporal cloud classifier. It is loaded only the first time it is required.
        """
        if self._multi_classifier is None:
            path = os.path.join(self.MODELS_FOLDER, self.MULTI_CLASSIFIER_NAME)
            self._multi_classifier = Booster(model_file=path)

        return self._multi_classifier

    @staticmethod
    def _run_prediction(classifier, features):
        """ Uses classifier object on given data
        """
        is_booster = isinstance(classifier, Booster)

        if is_booster:
            predict_method = classifier.predict
        else:
            # We assume it is a scikit-learn Estimator model
            predict_method = classifier.predict_proba

        prediction = execute_with_mp_lock(predict_method, features)

        if is_booster:
            return prediction
        return prediction[..., 1]

    def _scale_factors(self, reference_shape, bbox):
        """ Compute the resampling factors for height and width of the input array and sigma

        :param reference_shape: Tuple specifying height and width in pixels of high-resolution array
        :type reference_shape: (int, int)
        :param bbox: An EOPatch bounding box
        :type bbox: sentinelhub.BBox
        :return: Rescale factor for rows and columns
        :rtype: ((float, float), float)
        """
        res_x, res_y = bbox_to_resolution(bbox, width=reference_shape[1], height=reference_shape[0])

        if self.proc_resolution is None:
            pres_x, pres_y = res_x, res_y
        else:
            pres_x, pres_y = self.proc_resolution

        rescale = res_y / pres_y, res_x / pres_x
        sigma = 200 / (pres_x + pres_y)

        return rescale, sigma

    def _frame_indices(self, num_of_frames, target_idx):
        """ Returns frame indices within a given time window, with the target index relative to it
        """
        # Get reach
        nt_min = target_idx - self.max_proc_frames // 2
        nt_max = target_idx + self.max_proc_frames - self.max_proc_frames // 2

        # Shift reach
        shift = max(0, -nt_min) - max(0, nt_max - num_of_frames)
        nt_min += shift
        nt_max += shift

        # Get indices within range
        nt_min = max(0, nt_min)
        nt_max = min(num_of_frames, nt_max)
        nt_rel = target_idx - nt_min

        return nt_min, nt_max, nt_rel

    def _red_ssim(self, data_x, data_y, valid_mask, mu1, mu2, sigma1_2, sigma2_2, const1=1e-6, const2=1e-5):
        """ Slightly reduced (pre-computed) SSIM computation
        """
        # Increase precision and mask invalid regions
        valid_mask = valid_mask.astype(np.float64)
        data_x = data_x.astype(np.float64) * valid_mask
        data_y = data_y.astype(np.float64) * valid_mask

        # Init
        mu1_2 = mu1 * mu1
        mu2_2 = mu2 * mu2
        mu1_mu2 = mu1 * mu2

        sigma12 = cv2.GaussianBlur(
            (data_x * data_y).astype(np.float64), (0, 0), self.sigma, borderType=cv2.BORDER_REFLECT)
        sigma12 -= mu1_mu2

        # Formula
        tmp1 = 2. * mu1_mu2 + const1
        tmp2 = 2. * sigma12 + const2
        num = tmp1 * tmp2

        tmp1 = mu1_2 + mu2_2 + const1
        tmp2 = sigma1_2 + sigma2_2 + const2
        den = tmp1 * tmp2

        return np.divide(num, den)

    def _win_avg(self, data):
        """ Spatial window average
        """
        return cv2.GaussianBlur(data.astype(np.float64), (0, 0), self.sigma, borderType=cv2.BORDER_REFLECT)

    def _win_prevar(self, data):
        """ Incomplete spatial window variance
        """
        return cv2.GaussianBlur((data * data).astype(np.float64), (0, 0), self.sigma, borderType=cv2.BORDER_REFLECT)

    def _average(self, data):
        return cv2.filter2D(data.astype(np.float64), -1, self.avg_kernel, borderType=cv2.BORDER_REFLECT)

    def _dilate(self, data):
        return (cv2.dilate(data.astype(np.uint8), self.dil_kernel) > 0).astype(np.uint8)

    @staticmethod
    def _map_sequence(data, func2d):
        """ Iterate over time and band dimensions and apply a function to each slice.
        Returns a new array with the combined results.

        :param data: input array
        :type data: array of shape (timestamps, rows, columns, channels)
        :param func2d: Mapping function that is applied on each 2d image slice. All outputs must have the same shape.
        :type func2d: function (rows, columns) -> (new_rows, new_columns)
        """
        # Map over channel dimension on 3d tensor
        def func3d(dim):
            return map_over_axis(dim, func2d, axis=2)

        # Map over time dimension on 4d tensor
        def func4d(dim):
            return map_over_axis(dim, func3d, axis=0)

        output = func4d(data)

        return output

    def _average_all(self, data):
        """ Average over each spatial slice of data
        """
        if self.avg_kernel is not None:
            return self._map_sequence(data, self._average)

        return data

    def _dilate_all(self, data):
        """ Dilate over each spatial slice of data
        """
        if self.dil_kernel is not None:
            return self._map_sequence(data, self._dilate)

        return data

    def _ssim_stats(self, bands, is_data, win_avg, var, nt_rel):
        """ Calculate SSIM stats
        """
        ssim_max = np.empty((1, *bands.shape[1:]), dtype=np.float32)
        ssim_mean = np.empty_like(ssim_max)
        ssim_std = np.empty_like(ssim_max)

        bands_r = np.delete(bands, nt_rel, axis=0)
        win_avg_r = np.delete(win_avg, nt_rel, axis=0)
        var_r = np.delete(var, nt_rel, axis=0)

        n_frames = bands_r.shape[0]
        n_bands = bands_r.shape[-1]

        valid_mask = np.delete(is_data, nt_rel, axis=0) & is_data[nt_rel, ..., 0].reshape(
            1, *is_data.shape[1:-1], 1)

        for b_i in range(n_bands):
            local_ssim = []

            for t_j in range(n_frames):
                ssim_ij = self._red_ssim(bands[nt_rel, ..., b_i],
                                         bands_r[t_j, ..., b_i],
                                         valid_mask[t_j, ..., 0],
                                         win_avg[nt_rel, ..., b_i],
                                         win_avg_r[t_j, ..., b_i],
                                         var[nt_rel, ..., b_i],
                                         var_r[t_j, ..., b_i]
                                         )

                local_ssim.append(ssim_ij)

            local_ssim = np.ma.array(np.stack(local_ssim), mask=~valid_mask)

            ssim_max[0, ..., b_i] = np.ma.max(local_ssim, axis=0).data
            ssim_mean[0, ..., b_i] = np.ma.mean(local_ssim, axis=0).data
            ssim_std[0, ..., b_i] = np.ma.std(local_ssim, axis=0).data

        return ssim_max, ssim_mean, ssim_std

    def _do_single_temporal_cloud_detection(self, bands):
        """ Performs a cloud detection process on each scene separately
        """
        mono_proba = np.empty(np.prod(bands.shape[:-1]))
        img_size = np.prod(bands.shape[1:-1])

        n_times = bands.shape[0]

        for t_i in range(0, n_times, self.max_proc_frames):
            # Extract mono features
            nt_min = t_i
            nt_max = min(t_i + self.max_proc_frames, n_times)

            bands_t = bands[nt_min:nt_max]

            mono_features = bands_t.reshape(np.prod(bands_t.shape[:-1]), bands_t.shape[-1])

            mono_proba[nt_min * img_size:nt_max * img_size] = self._run_prediction(
                self.mono_classifier,
                mono_features
            )

        return mono_proba[..., np.newaxis]

    def _do_multi_temporal_cloud_detection(self, bands, is_data):
        """ Performs a cloud detection process on multiple scenes at once
        """
        multi_proba = np.empty(np.prod(bands.shape[:-1]))
        img_size = np.prod(bands.shape[1:-1])

        n_times = bands.shape[0]

        loc_mu = None
        loc_var = None

        prev_nt_min = None
        prev_nt_max = None

        for t_i in range(n_times):
            # Extract temporal window indices
            nt_min, nt_max, nt_rel = self._frame_indices(n_times, t_i)

            bands_t = bands[nt_min:nt_max]
            is_data_t = is_data[nt_min:nt_max]

            masked_bands = np.ma.array(bands_t, mask=~is_data_t.repeat(bands_t.shape[-1], axis=-1))

            # Add window averages and variances to local data
            if loc_mu is None or prev_nt_min != nt_min or prev_nt_max != nt_max:
                loc_mu, loc_var = self._update_batches(loc_mu, loc_var, bands_t, is_data_t)

            # Interweave and concatenate
            multi_features = self._extract_multi_features(bands_t, is_data_t, loc_mu, loc_var, nt_rel, masked_bands)

            multi_proba[t_i * img_size:(t_i + 1) * img_size] = self._run_prediction(
                self.multi_classifier,
                multi_features
            )

            prev_nt_min = nt_min
            prev_nt_max = nt_max

        return multi_proba[..., np.newaxis]

    def _update_batches(self, loc_mu, loc_var, bands_t, is_data_t):
        """ Updates window variance and mean values for a batch
        """
        # Add window averages and variances to local data
        if loc_mu is None:
            win_avg_bands = self._map_sequence(bands_t, self._win_avg)
            win_avg_is_data = self._map_sequence(is_data_t, self._win_avg)

            win_avg_is_data[win_avg_is_data == 0.] = 1.

            loc_mu = win_avg_bands / win_avg_is_data

            win_prevars = self._map_sequence(bands_t, self._win_prevar)
            win_prevars -= loc_mu * loc_mu

            loc_var = win_prevars

        else:
            win_avg_bands = self._map_sequence(
                bands_t[-1][None, ...], self._win_avg)
            win_avg_is_data = self._map_sequence(
                is_data_t[-1][None, ...], self._win_avg)

            win_avg_is_data[win_avg_is_data == 0.] = 1.

            loc_mu[:-1] = loc_mu[1:]
            loc_mu[-1] = (win_avg_bands / win_avg_is_data)[0]

            win_prevars = self._map_sequence(
                bands_t[-1][None, ...], self._win_prevar)
            win_prevars[0] -= loc_mu[-1] * loc_mu[-1]

            loc_var[:-1] = loc_var[1:]
            loc_var[-1] = win_prevars[0]

        return loc_mu, loc_var

    def _extract_multi_features(self, bands_t, is_data_t, loc_mu, loc_var, nt_rel, masked_bands):
        """ Extracts features for a batch
        """
        # Compute SSIM stats
        ssim_max, ssim_mean, ssim_std = self._ssim_stats(bands_t, is_data_t, loc_mu, loc_var, nt_rel)

        # Compute temporal stats
        temp_min = np.ma.min(masked_bands, axis=0).data[None, ...]
        temp_mean = np.ma.mean(masked_bands, axis=0).data[None, ...]

        # Compute difference stats
        t_all = len(bands_t)

        diff_max = (masked_bands[nt_rel][None, ...] - temp_min).data
        diff_mean = (masked_bands[nt_rel][None, ...] * (1. + 1. / (t_all - 1)) - t_all * temp_mean / (t_all - 1)).data

        # Interweave
        ssim_interweaved = np.empty((*ssim_max.shape[:-1], 3 * ssim_max.shape[-1]))
        ssim_interweaved[..., 0::3] = ssim_max
        ssim_interweaved[..., 1::3] = ssim_mean
        ssim_interweaved[..., 2::3] = ssim_std

        temp_interweaved = np.empty((*temp_min.shape[:-1], 2 * temp_min.shape[-1]))
        temp_interweaved[..., 0::2] = temp_min
        temp_interweaved[..., 1::2] = temp_mean

        diff_interweaved = np.empty((*diff_max.shape[:-1], 2 * diff_max.shape[-1]))
        diff_interweaved[..., 0::2] = diff_max
        diff_interweaved[..., 1::2] = diff_mean

        # Put it all together
        multi_features = np.concatenate(
            (bands_t[nt_rel][None, ...], loc_mu[nt_rel][None, ...],
             ssim_interweaved, temp_interweaved, diff_interweaved),
            axis=3
        )

        multi_features = multi_features.reshape(np.prod(multi_features.shape[:-1]), multi_features.shape[-1])

        return multi_features

    def execute(self, eopatch):
        """ Add selected features (cloud probabilities and masks) to an EOPatch instance.

        :param eopatch: Input `EOPatch` instance
        :return: `EOPatch` with additional features
        """
        data_feature = next(self.data_feature(eopatch))
        bands = eopatch[data_feature][..., self.band_indices].astype(np.float32)

        is_data_feature = next(self.is_data_feature(eopatch))
        is_data = eopatch[is_data_feature].astype(bool)

        original_shape = bands.shape[1: -1]
        scale_factors, self.sigma = self._scale_factors(original_shape, eopatch.bbox)

        is_data_sm = is_data
        # Downscale if specified
        if scale_factors is not None:
            bands = resize_images(bands.astype(np.float32), scale_factors=scale_factors)
            is_data_sm = resize_images(is_data.astype(np.uint8), scale_factors=scale_factors).astype(bool)

        mono_proba = None
        multi_proba = None
        mono_proba_feature, mono_mask_feature = self.mono_features
        multi_proba_feature, multi_mask_feature = self.multi_features

        # Run s2cloudless if needed
        if any([self.mask_feature, mono_mask_feature, mono_proba_feature]):
            mono_proba = self._do_single_temporal_cloud_detection(bands)
            mono_proba = mono_proba.reshape(*bands.shape[:-1], 1)

            # Upscale if necessary
            if scale_factors is not None:
                mono_proba = resize_images(mono_proba, new_size=original_shape)

            # Average over and threshold
            mono_mask = self._average_all(mono_proba) >= self.mono_threshold

        # Run SSIM-based multi-temporal classifier if needed
        if any([self.mask_feature, multi_mask_feature, multi_proba_feature]):
            multi_proba = self._do_multi_temporal_cloud_detection(bands, is_data_sm)
            multi_proba = multi_proba.reshape(*bands.shape[:-1], 1)

            # Upscale if necessary
            if scale_factors is not None:
                multi_proba = resize_images(multi_proba, new_size=original_shape)

            # Average over and threshold
            multi_mask = self._average_all(multi_proba) >= self.multi_threshold

        if mono_mask_feature is not None:
            mono_mask = self._dilate_all(mono_mask)
            eopatch.mask[mono_mask_feature] = (mono_mask * is_data).astype(bool)

        if multi_mask_feature is not None:
            multi_mask = self._dilate_all(multi_mask)
            eopatch.mask[multi_mask_feature] = (multi_mask * is_data).astype(bool)

        # Intersect
        if self.mask_feature:
            inter_mask = mono_mask & multi_mask
            inter_mask = self._dilate_all(inter_mask)
            eopatch[self.mask_feature] = (inter_mask * is_data).astype(bool)

        if mono_proba_feature is not None:
            eopatch.data[mono_proba_feature] = (mono_proba * is_data).astype(np.float32)

        if multi_proba_feature is not None:
            eopatch.data[multi_proba_feature] = (multi_proba * is_data).astype(np.float32)

        return eopatch


class AddMultiCloudMaskTask(CloudMaskTask):
    """ Temporary class for backward compatibility. Will raise a warning when used.
    """
    def __init__(self, *args, **kwargs):
        warnings.warn("AddMultiCloudMaskTask has been renamed to CloudMaskTask.", DeprecationWarning)
        super().__init__(*args, **kwargs)



"""
Utilities for cloud masking

Credits:
Copyright (c) 2017-2019 Matej Aleksandrov, Matej Batič, Andrej Burja(Sinergise)
Copyright (c) 2017-2019 Grega Milčinski, Matic Lubej, Devis Peresutti, Jernej Puc, Tomislav Slijepčević (Sinergise)
Copyright (c) 2017-2019 Jovan Višnjić, Anže Zupanc, Lojze Žust (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""

import numpy as np
import cv2


def map_over_axis(data, func, axis=0):
    """ Map function func over each slice along axis.
    If func changes the number of dimensions, mapping axis is moved to the front.

    Returns a new array with the combined results of mapping.

    :param data: input array
    :type data: np.array
    :param func: Mapping function that is applied on each slice. Outputs must have the same shape for every slice.
    :type func: function np.array -> np.array
    :param axis: Axis over which to map the function.
    :type axis: int

    :example:

    >>> data = np.ones((5,10,10))
    >>> func = lambda x: np.zeros((7,20))
    >>> res = map_over_axis(data,func,axis=0)
    >>> res.shape
    (5, 7, 20)
    """
    # Move axis to front
    data = np.moveaxis(data, axis, 0)

    res_mapped = [func(slice) for slice in data]
    res = np.stack(res_mapped)

    # Move axis back if number of dimensions stays the same
    if data.ndim == res.ndim:
        res = np.moveaxis(res, 0, axis)

    return res


def resize_images(data, new_size=None, scale_factors=None, anti_alias=True, interpolation='linear'):
    """Resizes the image(s) acording to given size or scale factors.

    To specify the new scale use one of `new_size` or `scale_factors` parameters.

    :param data: input image array
    :type data: numpy array with shape (timestamps, height, width, channels),
                (height, width, channels), or (height, width)
    :param new_size: New size of the data (height, width)
    :type new_size: (int, int)
    :param scale_factors: Factors (fy,fx) by which to resize the image
    :type scale_factors: (float, float)
    :param anti_alias: Use anti aliasing smoothing operation when downsampling. Default is True.
    :type anti_alias: bool
    :param interpolation: Interpolation method used for resampling.
                          One of 'nearest', 'linear', 'cubic'. Default is 'linear'.
    :type interpolation: string
    """
    inter_methods = {
        'nearest': cv2.INTER_NEAREST,
        'linear': cv2.INTER_LINEAR,
        'cubic': cv2.INTER_CUBIC
    }

    # Number of dimensions of input data
    ndims = data.ndim

    # Height and width axis indices for dimensionality
    height_width_axis = {2: (0, 1), 3: (0, 1), 4: (1, 2)}

    # Old height and width
    old_size = [data.shape[axis] for axis in height_width_axis[ndims]]

    if new_size is not None and scale_factors is None:
        if len(new_size) != 2 or any(not isinstance(value, int) for value in new_size):
            raise ValueError('new_size must be a pair of integers (height, width).')
        scale_factors = [new/old for old, new in zip(old_size, new_size)]
    elif scale_factors is not None and new_size is None:
        new_size = [int(size * factor) for size, factor in zip(old_size, scale_factors)]
    else:
        raise ValueError('Exactly one of the arguments new_size, scale_factors must be given.')

    if interpolation not in inter_methods:
        raise ValueError('Invalid interpolation method: %s' % interpolation)

    interpolation_method = inter_methods[interpolation]
    downscaling = scale_factors[0] < 1 or scale_factors[1] < 1

    def _resize2d(image):
        # Perform anti-alias smoothing if downscaling
        if downscaling and anti_alias:
            # Sigma computation based on skimage resize implementation
            sigmas = [((1/s) - 1)/2 for s in scale_factors]

            # Limit sigma values above 0
            sigma_y, sigma_x = [max(1e-8, sigma) for sigma in sigmas]
            image = cv2.GaussianBlur(image, (0, 0),
                                     sigmaX=sigma_x,
                                     sigmaY=sigma_y,
                                     borderType=cv2.BORDER_REFLECT)

        height, width = new_size
        resized = cv2.resize(image, (width, height), interpolation=interpolation_method)

        return resized

    _resize3d = lambda x: map_over_axis(x, _resize2d, axis=2)  # Map over channel dimension
    _resize4d = lambda x: map_over_axis(x, _resize3d, axis=0)  # Map over time dimension

    # Choose a resize method based on number of dimensions
    resize_methods = {2: _resize2d, 3: _resize3d, 4: _resize4d}

    resize_method = resize_methods[ndims]

    return resize_method(data)



"""
Public classes and functions of mask subpackage
"""

from .cloud_mask import AddMultiCloudMaskTask, CloudMaskTask
from .masking import AddValidDataMaskTask, MaskFeature
from .snow_mask import SnowMask, TheiaSnowMask
from .utilities import resize_images
from .mask_counting import ClassFrequencyTask

__version__ = '0.9.0'
